local options = MissionMinder.options;
local L = LibStub("AceLocale-3.0"):GetLocale("MissionMinder");
local DEBUG = MissionMinder.debug;

MissionMinder.planner = {};
local addon = MissionMinder.planner;

local RED_BIAS_PERCENTAGE = 0.10
local ORANGE_BIAS_PERCENTAGE = 0.50;
local WHITE_BIAS_PERCENTAGE = 1.00;

local LEVELUP_ARROW_QUALITIES = {
	[LE_ITEM_QUALITY_UNCOMMON]	= "loottoast-arrow-green",
	[LE_ITEM_QUALITY_RARE] = "loottoast-arrow-blue",
	[LE_ITEM_QUALITY_EPIC] = "loottoast-arrow-purple",
	[LE_ITEM_QUALITY_LEGENDARY] = "loottoast-arrow-orange"
}


local function createInlineTexture(texture)
	return texture and ("|T" .. texture .. ":0|t") or nil;
end

local function canBeSentOnMission(follower)
	return follower and follower.isCollected and follower.status ~= GARRISON_FOLLOWER_INACTIVE;
end

local function isOnMission(follower)
	return follower and follower.status == GARRISON_FOLLOWER_ON_MISSION;
end

local TRAIT_EXTRA_TRAINING = 80 -- whole party
local TRAIT_FAST_LEANER = 29 -- single follower
local TRAIT_HEARTHSTONE_PRO = 236 -- whole party

local expGainTraits = {
	[TRAIT_EXTRA_TRAINING] = 0.30,
	[TRAIT_FAST_LEANER] = 0.50,
	[TRAIT_HEARTHSTONE_PRO] = 0.35
}

local function ignoreTrait(traitID, hasBonusXPReward)
	if ( options.IgnoreXPGainTraits() and not hasBonusXPReward and expGainTraits[traitID]) then
		return true
	else
		return false
	end
end

local function followerWillMaxOnLevelUp(level, quality)
	local maxLevel = GarrisonMissionFrame.followerMaxLevel;
	local maxQuality = GarrisonMissionFrame.followerMaxQuality;

	local alreadyMaxLevel = level >= maxLevel
	local alreadyMaxQuality = quality >= maxQuality

	return (alreadyMaxQuality and level + 1 >= maxLevel) or (alreadyMaxLevel and quality + 1 >= maxQuality);
end



local function getMissionRewards(mission)
	local rewards = C_Garrison.GetMissionRewardInfo(mission.missionID);

	local t = {};
	for _, reward in pairs(rewards) do
		if(reward.followerXP) then
			local xp = reward.followerXP;
			if ( xp >= 1000 ) then
				xp = xp/1000 .. L["K"];
			end
			tinsert(t, "+" .. xp .. " " .. createInlineTexture(reward.icon));
		elseif (reward.itemID) then
			local itemTexture = select(10, GetItemInfo(reward.itemID));
			if(reward.quantity == 1) then
				tinsert(t, createInlineTexture(itemTexture));
			else
				tinsert(t, reward.quantity .. "x " .. createInlineTexture(itemTexture));
			end
		elseif(reward.currencyID) then
			if (reward.currencyID == 0) then -- gold
				tinsert(t, GetCoinTextureString(reward.quantity));
			else
				tinsert(t, reward.quantity .. " " .. createInlineTexture(reward.icon));
			end
		else
			-- because this is tooltip based, we ignore the value instead of writing the name
		end
	end

	return table.concat(t, ", ");
end

local function MissionHasBonusXPReward(missionID)
	local rewards = C_Garrison.GetMissionRewardInfo(missionID);
	for _, reward in pairs(rewards) do
		if(reward.followerXP) then
			return true;
		end
	end

	return false;
end

local timeLeft = {};
local inProgressMissions = {};
local function updateFollowerMissionTimeLeft(timeLeft)
	C_Garrison.GetInProgressMissions(inProgressMissions);

	if (inProgressMissions and #inProgressMissions > 0) then
		for _,mission in pairs(inProgressMissions) do
			for _,id in pairs(mission.followers) do
				-- Old values are overwritten, but if a follower is AVAILABLE any previous values will still be present.
				--Thats Okay because we only use this data if the follower is GARRISON_FOLLOWER_ON_MISSION.
				timeLeft[id] = {
					["missionID"] = mission.missionID,
					["remaining"] = mission.timeLeft;
				};
			end
		end
	end
end

local function getFollowerMissionBiasRGB(missionID,followerID)
	local bias = C_Garrison.GetFollowerBiasForMission(missionID, followerID);

	if ( bias == -1 ) then -- Impossible
		return 1, 0.1, 0.1;
	elseif ( bias < 0 ) then -- Hard
		return 1, 0.5, 0.25;
	else -- Normal
		return 1, 1, 1;
	end
end

local function levelComparator(left, right)
	if (left.iLevel ~= right.iLevel ) then
		return left.iLevel > right .iLevel;
	else
		return left.level > right.level;
	end
end

local function getMissionTimeleft(follower)
	if isOnMission(follower) then
		-- entry can be nil if follower is on a complete mission that has been viewed, but still needs to be looted.
		local entry = timeLeft[follower.followerID];
		if ( entry == nil or MissionMinder.IsMissionComplete(entry) ) then
			return L["completed"], nil, nil, nil; -- Default yellow
		else
			return string.format("(%s)", timeLeft[follower.followerID].remaining), 0.7, 0.7, 0.7; -- Gray
		end
	elseif (follower.status == GARRISON_FOLLOWER_WORKING) then
		return L["working"], 1, 0, 0; -- red
	else
		return L["available"], 0, 1, 0; -- Green
	end
end



local cache = {
	["isDirty"] = true,
	["abilities"] = {},
	["traits"] = {}
};


local function dirtyMissionCounterCache()
	cache.isDirty = true;
end

local function getFollowerCountersForMisison(followerID, missionID)
	if( cache.isDirty or not cache.abilities[missionID] or not cache.traits[missionID] ) then
		cache.abilities[missionID] = C_Garrison.GetBuffedFollowersForMission(missionID);
		cache.traits[missionID] = C_Garrison.GetFollowersTraitsForMission(missionID);
	end

	local abilities = cache.abilities[missionID];
	local traits = cache.traits[missionID];

	local hasBonusXP = MissionHasBonusXPReward(missionID);

	local counters = {};
	local hasNonFilterable = false
	if ( abilities and abilities[followerID] ) then
		for i = 1, #abilities[followerID] do
			local ability = abilities[followerID][i];
			counters[#counters + 1] = createInlineTexture(ability.icon);

			hasNonFilterable = true
		end
	end

	if ( traits and traits[followerID] ) then
		for i = 1, #traits[followerID] do
			local trait = traits[followerID][i];
			counters[#counters + 1] = createInlineTexture(trait.icon);

			if ( not ignoreTrait(trait.traitID, hasBonusXP) ) then
				hasNonFilterable = true
			end
		end
	end

	if ( hasNonFilterable ) then
		return counters;
	else
		return nil;
	end
end

local function createMissionTooltip(self, button)
	local mission = self.info;

	if(mission.inProgress) then -- Do not show tooltip on inProgress missions
		return;
	end

	--[===[@debug@
	GameTooltip:AddLine("MissionID = " .. mission.missionID);
	--@end-debug@]===]

	updateFollowerMissionTimeLeft(timeLeft);

	local followers = C_Garrison.GetFollowers(); -- GarrisonMissionFrame.FollowerList.followers
	local numFollowers = followers and #followers or 0;
	table.sort(followers, levelComparator);

	local counterAvailable = false;
	for i1 = 1, numFollowers do -- Necessary to walk all followers because we want to sort by level
		local follower = followers[i1];

		if ( canBeSentOnMission(follower) ) then
			local r1, g1, b1 = getFollowerMissionBiasRGB(mission.missionID, follower.followerID);

			local counters = getFollowerCountersForMisison(follower.followerID, mission.missionID);
			if ( counters and #counters > 0 ) then
				if ( not counterAvailable ) then
					GameTooltip:AddLine(" ");
					GameTooltip:AddLine(L["Counters"]);
					counterAvailable = true;
				end

				local level = follower.level
				if (mission.iLevel and mission.iLevel > 600 and follower.level == GARRISON_FOLLOWER_MAX_LEVEL) then
					level = follower.iLevel
				end

				local left = string.format("[%d] %s %s", level, follower.name, table.concat(counters, ""));
				local right, r2, g2, b2 = getMissionTimeleft(follower);
				GameTooltip:AddDoubleLine(left, right, r1, g1, b1, r2, g2, b2);
			end
		end
	end

	if ( not counterAvailable ) then
		GameTooltip:AddLine(" ");
		GameTooltip:AddLine(L["No counters available"]);
	end

	GameTooltip:Show(); -- resize the tooltip for the new lines
end

local function createFollowersTooltip(self)
	local follower = C_Garrison.GetFollowerInfo(self.id);
	if ( not follower.isCollected ) then
		return;
	end


	GameTooltip:SetOwner(self, "ANCHOR_NONE");
	GameTooltip:SetPoint("TOPRIGHT", self, "TOPLEFT", -10);
	local r, g, b = GetItemQualityColor(follower.quality);
	GameTooltip:SetText(follower.name, r, g, b);

	--[===[@debug@
	GameTooltip:AddLine("FollowerID = " .. follower.followerID);
	if ( GarrisonMissionFrame.MissionTab.MissionPage:IsShown() ) then
		local mission = GarrisonMissionFrame.MissionTab.MissionPage.missionInfo;
		local bias = C_Garrison.GetFollowerBiasForMission(mission.missionID, follower.followerID)
		GameTooltip:AddLine(string.format("Bias = %d%%", bias * 100));
	end
	--@end-debug@]===]

	if ( isOnMission(follower) ) then
		updateFollowerMissionTimeLeft(timeLeft);
		GameTooltip:AddLine(getMissionTimeleft(follower))
	end
	GameTooltip:AddLine(" ");

	local availableMissions = GarrisonMissionFrame.MissionTab.MissionList.availableMissions;
	C_Garrison.GetAvailableMissions(availableMissions);
	table.sort(availableMissions, levelComparator);

	local atleastOneMission = false;
	for _,mission in pairs(availableMissions) do
		local r, g, b = getFollowerMissionBiasRGB(mission.missionID, follower.followerID);

		local counters = getFollowerCountersForMisison(follower.followerID, mission.missionID);
		if ( counters and #counters > 0 ) then
			atleastOneMission = true;

			local level = mission.level
			if (mission.iLevel and mission.iLevel > 600) then
				level = mission.iLevel
			end

			local rewards = getMissionRewards(mission);
			local left = string.format("[%d] %s (%s)", level, mission.name, rewards);
			local right = table.concat(counters, "");
			GameTooltip:AddDoubleLine(left, right, r, g, b, 1, 1, 1);
		end
	end

	if ( not atleastOneMission ) then
		GameTooltip:AddLine(L["No counters available"]);
	end

	GameTooltip:Show();
	cache.isDirty = false;
end

local function biasColorFollowerLevels()
	local mission = GarrisonMissionFrame.MissionTab.MissionPage.missionInfo;
	local buttons = GarrisonMissionFrame.FollowerList.listScroll.buttons;

	for i = 1, #buttons do
		local button = buttons[i];
		local r, g, b = 1, 1, 1;

		if ( GarrisonMissionFrame.MissionTab.MissionPage:IsShown() and not GarrisonMissionFrame.FollowerTab:IsShown()
				and mission and button.isCollected) then
			r, g, b = getFollowerMissionBiasRGB(mission.missionID, button.id);
		end

		button.PortraitFrame.Level:SetTextColor(r, g, b)
	end
end

local function getPotentialMissionXPGains(mission, follower)
	if ( not mission or not follower ) then
		return 0
	end

	local _, base = C_Garrison.GetMissionInfo(mission.missionID);

	local bonus = 0;
	for _, reward in pairs(C_Garrison.GetMissionRewardInfo(mission.missionID)) do
		if(reward.followerXP) then
			bonus = bonus + reward.followerXP
		end
	end

	local _, _, _, _, partyBuffs, _, modify = C_Garrison.GetPartyMissionInfo(mission.missionID);
	local gain = 0;
	for i = 1, #partyBuffs do
		if ( partyBuffs[i] ~= TRAIT_FAST_LEANER ) then
			gain = gain + (expGainTraits[partyBuffs[i]] or 0);
		end
	end

	if ( follower.abilities ) then
		for i = 1, #follower.abilities do
			local ability = follower.abilities[i]
			if ( ability.isTrait and ability.id == TRAIT_FAST_LEANER ) then
				gain = gain + (expGainTraits[ability.id] or 0);
			end
		end
	end

	local bias = C_Garrison.GetFollowerBiasForMission(mission.missionID, follower.followerID);
	if ( mission.iLevel and mission.iLevel > 600 ) then
		if ( bias <= -1 ) then
			bias = ORANGE_BIAS_PERCENTAGE
		else
			bias = WHITE_BIAS_PERCENTAGE
		end
	else
		if ( bias <= -1 ) then
			bias = RED_BIAS_PERCENTAGE;
		elseif ( bias < 0) then
			bias = ORANGE_BIAS_PERCENTAGE
		else
			bias = WHITE_BIAS_PERCENTAGE;
		end
	end

	return bias * (base + modify + (bonus * (1 + gain)));
end

local function getActualMissionXPGains(potentialXP, follower)
	if( potentialXP > 0 ) then
		-- Maxed followers don't get any XP
		if ( not GarrisonMissionFrame_GetFollowerNextLevelXP(follower.level, follower.quality) ) then
			return 0;
		end

		-- Follower expected to level and max, only show limited amount of XP they would gain
		local maxXP = GarrisonMissionFrame_GetFollowerNextLevelXP(follower.level, follower.quality)
		if ( maxXP ) then
			local remainingXP = maxXP - follower.xp;
			if ( potentialXP >= remainingXP and followerWillMaxOnLevelUp(follower.level, follower.quality) ) then
				return remainingXP;
			end
		end
	end

	return potentialXP;
end

local function updateLevelUp(self)
	local missionPage = GarrisonMissionFrame.MissionTab.MissionPage
	local mission = missionPage.missionInfo;
	local follower = self.info

	if ( mission and missionPage:IsShown() and not GarrisonMissionFrame.FollowerTab:IsShown() and canBeSentOnMission(follower)) then
		local maxXP = GarrisonMissionFrame_GetFollowerNextLevelXP(follower.level, follower.quality);
		local missionXPGains = getPotentialMissionXPGains(mission, follower);

		if ( maxXP and follower.xp + missionXPGains >= maxXP ) then
			self.levelUp:SetAtlas(LEVELUP_ARROW_QUALITIES[follower.quality], true);
			self.levelUp:Show();
		else
			self.levelUp:Hide();
		end
	else
		self.levelUp:Hide();
	end
end

local function updateXPGains(self)
	local missionPage = GarrisonMissionFrame.MissionTab.MissionPage
	local mission = missionPage.missionInfo;
	local follower = self.info

	local potentialXP = getPotentialMissionXPGains(mission, follower)
	local missionXPGains = getActualMissionXPGains(potentialXP, follower);
	if (missionXPGains > 0) then
		self.XPGain:SetFormattedText(XP_GAIN, BreakUpLargeNumbers(math.ceil(missionXPGains)));
		self.XPGain:Show();
	else
		self.XPGain:Hide();
	end
end

local function updateXpRemaining(self)
	local follower = self.info

	if ( follower and follower.isCollected ) then
		local maxXP = GarrisonMissionFrame_GetFollowerNextLevelXP(follower.level, follower.quality);
		if ( maxXP ) then
			self.xpRemaining:SetFormattedText(GARRISON_REWARD_XP_FORMAT, BreakUpLargeNumbers(maxXP - follower.xp));
		else
			self.xpRemaining:SetText(nil);
		end
	else
		self.xpRemaining:SetText(nil);
	end
end

local function updateFollowerList()
	local buttons = GarrisonMissionFrame.FollowerList.listScroll.buttons;
	for i = 1, #buttons do
		updateLevelUp(buttons[i]);
		updateXpRemaining(buttons[i]);
	end
end

local function updateParty()
	local partyFollowers = GarrisonMissionFrame.MissionTab.MissionPage.Followers;
	for i = 1, #partyFollowers do
		updateLevelUp(partyFollowers[i]);
		updateXPGains(partyFollowers[i]);
	end
end

local function hideMissionCountersTooltip(self)
	GameTooltip:Hide();
end


local singleclickTimer = nil;
-- Override the standard OnClick event to delay just long enough to determine if it is a DoubleClick.
--If it is a double click then the OnClick event will not fire. If it isn't a double click then the
--OnClick event (which will run via timer) will still run; albeit slightly delayed.
local function followerSingleClick(self, button)
	if(singleclickTimer) then
		singleclickTimer:Cancel();
	end

	-- Only delay OnClick in very specific cases
	if ( self.isCollected and button == "LeftButton" ) then
		if ( GarrisonMissionFrame.MissionTab.MissionPage:IsShown() and not GarrisonMissionFrame.FollowerTab:IsShown() ) then
			singleclickTimer = C_Timer.NewTimer(0.2, function()
				addon.GarrisonFollowerListButton_OnClick(self, button);
			end);
			return;
		end
	end

	addon.GarrisonFollowerListButton_OnClick(self, button);
end

-- Handle OnDoubleClick event and disable the single click timer
local function assignFollowerToMission(self, button)
	if(singleclickTimer) then
		singleclickTimer:Cancel();
	end

	-- Hide the rightclick menu
	GarrisonFollowerOptionDropDown.followerID = nil;
	CloseDropDownMenus();

	if ( self.isCollected and self.info.status == nil and button == "LeftButton" ) then
		local mission = GarrisonMissionFrame.MissionTab.MissionPage.missionInfo;
		if ( mission and GarrisonMissionFrame.MissionTab.MissionPage:IsShown()
					and not GarrisonMissionFrame.FollowerTab:IsShown() ) then
			DEBUG("assignFollowerToMission(mission=%s, follower=%s)", mission.missionID, self.id);
			GarrisonMissionPage_AddFollower(self.id);
		end
	end
end

local function disableTimer()
	if(singleclickTimer) then
		singleclickTimer:Cancel()
		singleclickTimer = nil
	end
end

local followerDragSource = nil;
local function setDragSource(self, button)
	if (GarrisonFollowerPlacer:IsShown()) then
		DEBUG("SetDragSource(%s)", tostring(self));
		followerDragSource = self;
	end
end

local function clearDragSource()
DEBUG("ClearDragSource()");
	followerDragSource = nil;
end

local function getFollowerFrameIndex(self)
	local partyFollowers = GarrisonMissionFrame.MissionTab.MissionPage.Followers;
	for i = 1, #partyFollowers do
		if ( partyFollowers[i] == self) then
			return i
		end
	end

	return 0;
end

local function setFollower(destFrame, destInfo, ...)
	DEBUG("SetFollower(%s, %s) DragSource=%s", tostring(destFrame), destInfo.followerID, tostring(followerDragSource));
	if ( followerDragSource ) then
		local srcFrame = followerDragSource;
		local srcInfo = destFrame.info
		srcIndex = getFollowerFrameIndex(srcFrame);
		destIndex = getFollowerFrameIndex(destFrame);

		DEBUG("srcIndex=%d, destIndex=%d", srcIndex, destIndex)

		if (srcIndex > destIndex ) then
			-- We always want to assign the "lower" frame first to keep the follower order consistent
			--There are some internals to garrison missions that can cause weirdness for the user.
			--like hidden enemy modifiers on [615] Auchindoom!
			srcFrame, destFrame = destFrame, srcFrame
			srcInfo, destInfo = destInfo, srcInfo
		elseif (srcIndex < destIndex ) then
			-- Must first unassign dest follower
			GarrisonMissionPage_ClearFollower(destFrame);
		else -- Targetting the same frame, only do one set (destination)
			srcInfo = nil;
		end

		if(srcInfo) then
			DEBUG("src SetFollower(%s, %s)", tostring(srcFrame), srcInfo.followerID);
			addon.GarrisonMissionPage_SetFollower(srcFrame, srcInfo, ...)
		end
		if (destInfo) then
			DEBUG("dest SetFollower(%s, %s)", tostring(destFrame), destInfo.followerID);
			addon.GarrisonMissionPage_SetFollower(destFrame, destInfo, ...)
		end
	else
		addon.GarrisonMissionPage_SetFollower(destFrame, destInfo, ...)
	end
end

local function ShowBonusXPAmount(self, rewards, numRewards)
	if (numRewards > 0 and rewards) then
		local index = 1;
		for id, reward in pairs(rewards) do
			if (self.Rewards[index]) then -- Should never be nil, but if it ever is we ignore it
				local Reward = self.Rewards[index];
				if ( reward.followerXP and reward.followerXP > 0 ) then
					Reward.Quantity:SetText(BreakUpLargeNumbers(reward.followerXP));
					Reward.Quantity:Show();
				elseif (reward.currencyID == 0) then -- gold
					Reward.Quantity:SetText(math.floor(reward.quantity / 10000));
					Reward.Quantity:Show();
				end
			end

			index = index + 1;
		end
	end
end

function addon.initialize()
	local missionButtons = GarrisonMissionFrame.MissionTab.MissionList.listScroll.buttons;
	for i = 1, #missionButtons do
		missionButtons[i]:HookScript("OnEnter", createMissionTooltip);
	end

	addon.GarrisonFollowerListButton_OnClick = GarrisonFollowerListButton_OnClick;

	local followerButtons = GarrisonMissionFrame.FollowerList.listScroll.buttons;
	for i = 1, #followerButtons do
		followerButtons[i]:SetScript("OnEnter", createFollowersTooltip);
		followerButtons[i]:SetScript("OnLeave", hideMissionCountersTooltip);

		-- Allow double click to assign followers to mission
		followerButtons[i]:SetScript("OnClick", followerSingleClick);
		followerButtons[i]:SetScript("OnDoubleClick", assignFollowerToMission);

		-- Clear drag source when dragging from follower list
		followerButtons[i]:HookScript("OnDragStart", clearDragSource)

		local levelUp = followerButtons[i].PortraitFrame:CreateTexture();
		levelUp:SetPoint("BOTTOM", followerButtons[i].PortraitFrame, "BOTTOMLEFT")
		levelUp:Hide();
		followerButtons[i].levelUp = levelUp

		local xpRemaining = followerButtons[i]:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall");
		xpRemaining:SetPoint("BOTTOMRIGHT", followerButtons[i].BG);
		xpRemaining:SetText(nil);
		xpRemaining:Show();
		followerButtons[i].xpRemaining = xpRemaining;
	end

	-- Color follower levels by bias
	hooksecurefunc("GarrisonFollowerList_Update", biasColorFollowerLevels)
	hooksecurefunc("GarrisonFollowerList_Update", updateFollowerList)

	-- Make sure we update the list after showing the mission page. This is important because
	--the mission info is not initially available when the list first shows
	hooksecurefunc("GarrisonMissionPage_ShowMission", biasColorFollowerLevels);

	-- Disable time on MissionPage:OnHide
	GarrisonMissionFrame.MissionTab.MissionPage:HookScript("OnHide", disableTimer);

	-- Dragging from party followers sets the drag source
	local partyFollowers = GarrisonMissionFrame.MissionTab.MissionPage.Followers;
	for i = 1, #partyFollowers do
		partyFollowers[i]:HookScript("OnDragStart", setDragSource);
		partyFollowers[i]:HookScript("OnReceiveDrag", clearDragSource);

		local levelUp = partyFollowers[i].PortraitFrame:CreateTexture();
		levelUp:SetPoint("BOTTOM", partyFollowers[i].PortraitFrame, "BOTTOMLEFT")
		levelUp:Hide();
		partyFollowers[i].levelUp = levelUp

		local xpGain = partyFollowers[i]:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall")
		xpGain:SetPoint("TOP", partyFollowers[i], "BOTTOM")
		xpGain:SetTextColor(0, 1, 0)
		xpGain:SetText(nil)
		xpGain:Show()
		partyFollowers[i].XPGain = xpGain
	end

	hooksecurefunc("GarrisonMissionPage_SetFollower", updateParty);
	hooksecurefunc("GarrisonMissionPage_ClearFollower", updateParty);

	-- Override the SetFollower method
	addon.GarrisonMissionPage_SetFollower = GarrisonMissionPage_SetFollower
	GarrisonMissionPage_SetFollower = setFollower
	-- Clear DragSource on PlacerFrame click
	GarrisonFollowerPlacerFrame:HookScript("OnClick", clearDragSource);

	hooksecurefunc("GarrisonMissionButton_SetRewards", ShowBonusXPAmount);
end


local events = MissionMinder.events;
events:RegisterEvent("GARRISON_FOLLOWER_LIST_UPDATE", dirtyMissionCounterCache);
events:RegisterEvent("GARRISON_FOLLOWER_XP_CHANGED", dirtyMissionCounterCache);
events:RegisterEvent("GARRISON_FOLLOWER_REMOVED", dirtyMissionCounterCache);
